import re

from django.conf import settings
from django.utils.html import escape, html_safe, mark_safe


TEMPLATES = {
    "opengraph": '<meta property="{name}" content="{content}">',
    "meta": '<meta name="{name}" content="{content}">',
    "link": '<link rel="{name}" href="{content}">',
}
TAGS = [
    ("opengraph", "og:description", "description"),
    ("opengraph", "og:image", "image"),
    ("opengraph", "og:image:width", "image:width"),
    ("opengraph", "og:image:height", "image:height"),
    ("opengraph", "og:image:alt", "image:alt"),
    ("opengraph", "og:site_name", "site_name"),
    ("opengraph", "og:title", "title"),
    ("opengraph", "og:type", "type"),
    ("opengraph", "og:url", "url"),
    ("meta", "description", "description"),
    ("meta", "author", "author"),
    ("meta", "robots", "robots"),
    ("link", "canonical", "canonical"),
]
URLS = {"canonical", "image", "url"}


_attribute_escapes = {
    ord("<"): "&lt;",
    ord(">"): "&gt;",
    ord("&"): "&amp;",
}


def escape_attribute(s):
    """
    Preserve single quotes etc. in attribute values if it is safe. Some user
    agents do not seem to interpret HTML entities in Open Graph values
    correctly.
    """
    if re.match(r"^[-\w\s_'&:;.,/()]+$", s, re.I | re.U):
        return s.translate(_attribute_escapes)
    return escape(s)


@html_safe
class MetaTags(dict):
    def __str__(self):
        """
        Return a safe HTML representation of the meta tag dictionary
        """
        uri = self.get("build_absolute_uri", lambda x: x)
        return mark_safe(
            "".join(
                TEMPLATES[template].format(
                    name=name,
                    content=escape_attribute(
                        uri(str(self[key])) if key in URLS else str(self[key])
                    ),
                )
                for template, name, key in TAGS
                if self.get(key)
            )
        )

    def update(self, other):
        if other:
            for key, value in other.items():
                if value is None:
                    self.pop(key, None)
                elif value:
                    self[key] = value
        return self

    def add(self, *dicts_or_meta_mixins, **kwargs):
        for thing in dicts_or_meta_mixins:
            self.update(thing.meta_dict() if hasattr(thing, "meta_dict") else thing)
        self.update(kwargs)
        return self


def meta_tags(objects=(), *, request, defaults=None, **kwargs):
    """
    Return a dictionary containing meta tag information. Keys include
    (but are not limited to):

    - ``title``
    - ``description``
    - ``image`` (Full URL)
    - ``canonical`` (Full URL)
    - ``url`` (Full URL, overridden using ``canonical``)
    - ``author``
    - ``robots``

    Data is taken in order from the following sources:

    - Any keyword arguments to this function (items with a value of ``None``
      are discarded)
    - Any ``MetaMixin`` instances passed as an iterable
    - The ``defaults`` argument, if given
    - The ``META_TAGS`` setting, if available

    The recommended set of keys for ``META_TAGS`` is ``site_name``,
    ``title``, ``description`` and ``image``. See
    `The Open Graph protocol <http://ogp.me/>`_ for additional details.
    """
    return MetaTags(
        build_absolute_uri=request.build_absolute_uri,
        type="website",
        url=request.get_full_path(),
    ).add(getattr(settings, "META_TAGS", None), defaults, *reversed(objects), **kwargs)



from django.db import models
from django.utils.translation import gettext_lazy as _
from imagefield.fields import ImageField


class MetaMixin(models.Model):
    meta_title = models.CharField(
        _("title"),
        max_length=200,
        blank=True,
        default="",
        help_text=_("Used for Open Graph and other meta tags."),
    )
    meta_description = models.TextField(
        _("description"),
        blank=True,
        default="",
        help_text=_("Override the description for this page."),
    )
    meta_image = ImageField(
        _("image"),
        blank=True,
        default="",
        auto_add_fields=True,
        upload_to="meta/%Y/%m",
        help_text=_("Set the Open Graph image."),
        formats={"opengraph": ("default", ("crop", (1200, 630)))},
    )
    meta_image_alternative_text = models.CharField(
        _("alternative text"),
        blank=True,
        default="",
        help_text=_("Describe the contents, e.g. for screenreaders."),
        max_length=1000,
    )
    meta_canonical = models.URLField(
        _("canonical URL"),
        blank=True,
        default="",
        help_text=_("If you need this you probably know."),
    )
    meta_author = models.CharField(
        _("author"),
        blank=True,
        default="",
        help_text=_("Override the author meta tag."),
        max_length=200,
    )
    meta_robots = models.CharField(
        _("robots"),
        blank=True,
        default="",
        help_text=_("Override the robots meta tag."),
        max_length=200,
    )

    class Meta:
        abstract = True

    @classmethod
    def admin_fieldset(cls, **kwargs):
        cfg = {
            "fields": (
                "meta_title",
                "meta_description",
                "meta_image",
                "meta_image_ppoi",
                "meta_image_alternative_text",
                "meta_canonical",
                "meta_author",
                "meta_robots",
            ),
            "classes": ("tabbed",),
        }
        cfg.update(kwargs)
        return (_("Meta tags"), cfg)

    def meta_dict(self):
        ctx = {
            "title": self.meta_title or getattr(self, "title", ""),
            "description": self.meta_description,
            "canonical": self.meta_canonical,
            # Override URL if canonical is set to a non-empty value (the empty
            # string will be skipped when merging this dictionary)
            "url": self.meta_canonical,
            "author": self.meta_author,
            "robots": self.meta_robots,
        }
        ctx.update(self.meta_images_dict())
        return ctx

    def meta_images_dict(self):
        if self.meta_image:
            return {
                "image": str(self.meta_image.opengraph),
                "image:width": 1200,
                "image:height": 630,
                "image:alt": self.meta_image_alternative_text,
            }

        elif getattr(self, "image", None):
            return {"image": self.image.url}

        return {"image": ""}
